__all__ = ['Host', 'TCPServiceClient', 'TCPService', 'HTTPService', 'HTTPServiceClient', 'api', 'request', 'subscribe',
           'publish', 'xsubscribe', 'get', 'post', 'head', 'put', 'patch', 'delete', 'options', 'trace',
           'RequestException', 'Response', 'Request', 'log', 'setup_logging', 'apideprecated',
           'TrellioServiceException', 'TrellioServiceError', 'ConfigHandler', 'ManagementCommand', 'HTTPView',
           'TCPView', 'ManagementRegistry', 'InvalidCMDArguments', 'execute_from_command_line', 'Publisher',
           'Subscriber']

import logging

logger = logging.getLogger(__name__)

try:
    import asyncio
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    logger.warning('uvloop is not install, event loop will be set to default asyncio loop')

from .conf_manager import *
from .exceptions import RequestException, TrellioServiceError, TrellioServiceException  # noqa
from .host import Host  # noqa
from .management import *
from .pubsub import Publisher, Subscriber
from .services import (TCPService, HTTPService, HTTPServiceClient, TCPServiceClient)  # noqa
from .services import (api, request, subscribe, publish, xsubscribe, apideprecated)  # noqa
from .services import (get, post, head, put, patch, delete, options, trace)  # noqa
from .utils import log  # noqa
from .utils.log import setup_logging  # noqa
from .views import HTTPView, TCPView
from .wrappers import Response, Request  # noqa

__version__ = '1.1.33rc2'



from ..wrappers import Response


class Borg(object):
    __shared_state = dict()

    def __init__(self):
        self.__dict__ = self.__shared_state


class Singleton(object):
    _instance = None
    _init_ran = False

    def has_inited(self):
        return self._init_ran

    def init_done(self):
        self._init_ran = True

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(Singleton, cls).__new__(cls, *args, **kwargs)
        return cls._instance


def default_preflight_response(request):
    headers = {'Access-Control-Allow-Origin': '*',
               'Access-Control-Allow-Methods': 'GET,POST,PUT,DELETE',
               'Access-Control-Allow-Headers': 'Access-Control-Allow-Headers, Origin,Accept, X-Requested-With, Content-Type, Authorization, Access-Control-Request-Method, Access-Control-Request-Headers',
               'Access-Control-Allow-Credentials': 'true'}
    return Response(status=204, headers=headers)



import asyncio
import logging
import os
import signal
import warnings
from functools import partial

from aiohttp.web import Application

from .bus import TCPBus
from .protocol_factory import get_trellio_protocol
from .pubsub import Publisher, Subscriber
from .registry_client import RegistryClient
from .services import HTTPService, TCPService
from .signals import ServiceReady
from .utils.decorators import deprecated
from .utils.log import setup_logging
from .utils.stats import Stats, Aggregator


class Host:
    """Serves as a static entry point and provides the boilerplate required to host and run a trellio Service.

    Example::

        Host.configure('SampleService')
        Host.attachService(SampleHTTPService())
        Host.run()

    """
    registry_host = None
    registry_port = None
    pubsub_host = None
    pubsub_port = None
    host_name = None
    service_name = None
    http_host = None
    http_port = None
    tcp_host = None
    tcp_port = None
    ssl_context = None
    ronin = False  # If true, the trellio service runs solo without a registry

    _host_id = None
    _tcp_service = None
    _http_service = None
    _publisher = None
    _subscribers = []
    _tcp_views = []
    _http_views = []
    _logger = logging.getLogger(__name__)
    _smtp_handler = None

    @classmethod
    def configure(cls, host_name: str = '', service_name: str = '', service_version='',
                  http_host: str = '127.0.0.1', http_port: int = 8000,
                  tcp_host: str = '127.0.0.1', tcp_port: int = 8001, ssl_context=None,
                  registry_host: str = "0.0.0.0", registry_port: int = 4500,
                  pubsub_host: str = "0.0.0.0", pubsub_port: int = 6379, ronin: bool = False):
        """ A convenience method for providing registry and pubsub(redis) endpoints

        :param host_name: Used for process name
        :param registry_host: IP Address for trellio-registry; default = 0.0.0.0
        :param registry_port: Port for trellio-registry; default = 4500
        :param pubsub_host: IP Address for pubsub component, usually redis; default = 0.0.0.0
        :param pubsub_port: Port for pubsub component; default= 6379
        :return: None
        """
        Host.host_name = host_name
        Host.service_name = service_name
        Host.service_version = str(service_version)
        Host.http_host = http_host
        Host.http_port = http_port
        Host.tcp_host = tcp_host
        Host.tcp_port = tcp_port
        Host.registry_host = registry_host
        Host.registry_port = registry_port
        Host.pubsub_host = pubsub_host
        Host.pubsub_port = pubsub_port
        Host.ssl_context = ssl_context
        Host.ronin = ronin

    @classmethod
    def get_http_service(cls):
        return cls._http_service

    @classmethod
    def get_tcp_service(cls):
        return cls._tcp_service

    @classmethod
    def get_tcp_clients(cls):
        tcp_service = cls.get_tcp_service()
        if tcp_service:
            return tcp_service.clients

    @classmethod
    def get_publisher(cls):
        return cls._publisher

    @classmethod
    def get_subscribers(cls):
        return cls._subscribers

    @classmethod
    def get_tcp_views(cls):
        return cls._tcp_views

    @classmethod
    def get_http_views(cls):
        return cls._http_views

    @classmethod
    @deprecated
    def attach_service(cls, service):
        """ Allows you to attach one TCP and one HTTP service

        deprecated:: 2.1.73 use http and tcp specific methods
        :param service: A trellio TCP or HTTP service that needs to be hosted
        """
        if isinstance(service, HTTPService):
            cls._http_service = service
        elif isinstance(service, TCPService):
            cls._tcp_service = service
        else:
            cls._logger.error('Invalid argument attached as service')
        cls._set_bus(service)

    @classmethod
    def attach_http_service(cls, http_service: HTTPService):
        """ Attaches a service for hosting
        :param http_service: A HTTPService instance
        """
        if cls._http_service is None:
            cls._http_service = http_service
            cls._set_bus(http_service)
        else:
            warnings.warn('HTTP service is already attached')

    @classmethod
    def attach_tcp_service(cls, tcp_service: TCPService):
        """ Attaches a service for hosting
        :param tcp_service: A TCPService instance
        """
        if cls._tcp_service is None:
            cls._tcp_service = tcp_service
            cls._set_bus(tcp_service)
        else:
            warnings.warn('TCP service is already attached')

    @classmethod
    def attach_http_views(cls, http_views: list):
        views_instances = []
        for view_class in http_views:
            instance = view_class()
            instance.host = Host
            views_instances.append(instance)
        cls._http_views.extend(views_instances)

    @classmethod
    def attach_tcp_views(cls, tcp_views: list):
        views_instances = []
        for view_class in tcp_views:
            instance = view_class()
            instance.host = Host
            views_instances.append(instance)
        cls._tcp_views.extend(views_instances)

    @classmethod
    def attach_publisher(cls, publisher: Publisher):
        if cls._publisher is None:
            cls._publisher = publisher
        else:
            warnings.warn('Publisher is already attached')

    @classmethod
    def attach_subscribers(cls, subscribers: list):
        if all([isinstance(subscriber, Subscriber) for subscriber in subscribers]):
            if not cls._subscribers:
                cls._subscribers = subscribers
            else:
                warnings.warn('Subscribers are already attached')

    @classmethod
    def run(cls):
        """ Fires up the event loop and starts serving attached services
        """
        if cls._tcp_service or cls._http_service or cls._http_views or cls._tcp_views:
            cls._set_host_id()
            cls._setup_logging()
            cls._set_process_name()
            cls._set_signal_handlers()
            cls._start_pubsub()
            cls._start_server()
        else:
            cls._logger.error('No services to host')

    @classmethod
    def _set_process_name(cls):
        from setproctitle import setproctitle
        setproctitle('trellio_{}_{}'.format(cls.host_name, cls._host_id))

    @classmethod
    def _stop(cls, signame: str):
        cls._logger.info('\ngot signal {} - exiting'.format(signame))
        asyncio.get_event_loop().stop()

    @classmethod
    def _set_signal_handlers(cls):
        asyncio.get_event_loop().add_signal_handler(getattr(signal, 'SIGINT'), partial(cls._stop, 'SIGINT'))
        asyncio.get_event_loop().add_signal_handler(getattr(signal, 'SIGTERM'), partial(cls._stop, 'SIGTERM'))

    @classmethod
    def _create_tcp_server(cls):
        if cls._tcp_service:
            ssl_context = cls._tcp_service.ssl_context
            host_ip, host_port = cls._tcp_service.socket_address
            task = asyncio.get_event_loop().create_server(partial(get_trellio_protocol, cls._tcp_service.tcp_bus),
                                                          host_ip, host_port, ssl=ssl_context)
            result = asyncio.get_event_loop().run_until_complete(task)
            return result

    @classmethod
    def _create_http_server(cls):
        if cls._http_service or cls._http_views:
            host_ip, host_port = cls.http_host, cls.http_port
            ssl_context = cls.ssl_context
            handler = cls._make_aiohttp_handler()
            task = asyncio.get_event_loop().create_server(handler, host_ip, host_port, ssl=ssl_context)
            return asyncio.get_event_loop().run_until_complete(task)

    @classmethod
    def _make_aiohttp_handler(cls):
        app = Application(loop=asyncio.get_event_loop())

        if cls._http_service:
            for each in cls._http_service.__ordered__:
                # iterate all attributes in the service looking for http endpoints and add them
                fn = getattr(cls._http_service, each)
                if callable(fn) and getattr(fn, 'is_http_method', False):
                    for path in fn.paths:
                        app.router.add_route(fn.method, path, fn)
                        if cls._http_service.cross_domain_allowed:
                            # add an 'options' for this specific path to make it CORS friendly
                            app.router.add_route('options', path, cls._http_service.preflight_response)

        for view in cls._http_views:
            for each in view.__ordered__:
                fn = getattr(view, each)
                if callable(fn) and getattr(fn, 'is_http_method', False):
                    for path in fn.paths:
                        app.router.add_route(fn.method, path, fn)
                        if view.cross_domain_allowed:
                            # add an 'options' for this specific path to make it CORS friendly
                            app.router.add_route('options', path, view.preflight_response)

        handler = app.make_handler(access_log=cls._logger)
        return handler

    @classmethod
    def _set_host_id(cls):
        from uuid import uuid4
        cls._host_id = uuid4()

    @classmethod
    def _start_server(cls):
        tcp_server = cls._create_tcp_server()
        http_server = cls._create_http_server()
        if not cls.ronin:
            if cls._tcp_service:
                asyncio.get_event_loop().run_until_complete(cls._tcp_service.tcp_bus.connect())
                # if cls._http_service:
                #     asyncio.get_event_loop().run_until_complete(cls._http_service.tcp_bus.connect())
        if tcp_server:
            cls._logger.info('Serving TCP on {}'.format(tcp_server.sockets[0].getsockname()))
        if http_server:
            cls._logger.info('Serving HTTP on {}'.format(http_server.sockets[0].getsockname()))
        cls._logger.info("Event loop running forever, press CTRL+C to interrupt.")
        cls._logger.info("pid %s: send SIGINT or SIGTERM to exit." % os.getpid())
        cls._logger.info("Triggering ServiceReady signal")
        asyncio.get_event_loop().run_until_complete(ServiceReady._run())
        try:
            asyncio.get_event_loop().run_forever()
        except Exception as e:
            print(e)
        finally:
            if tcp_server:
                tcp_server.close()
                asyncio.get_event_loop().run_until_complete(tcp_server.wait_closed())

            if http_server:
                http_server.close()
                asyncio.get_event_loop().run_until_complete(http_server.wait_closed())

            asyncio.get_event_loop().close()

    @classmethod
    def _start_pubsub(cls):
        if not cls.ronin:
            if cls._publisher:
                asyncio.get_event_loop().run_until_complete(cls._publisher.create_pubsub_handler())

        for subscriber in cls._subscribers:
            asyncio.get_event_loop().run_until_complete(subscriber.create_pubsub_handler())
            asyncio.async(subscriber.register_for_subscription())

    @classmethod
    def _set_bus(cls, service):
        registry_client = RegistryClient(asyncio.get_event_loop(), cls.registry_host, cls.registry_port)
        tcp_bus = TCPBus(registry_client)
        registry_client.conn_handler = tcp_bus
        # pubsub_bus = PubSubBus(cls.pubsub_host, cls.pubsub_port, registry_client)  # , cls._tcp_service._ssl_context)
        registry_client.bus = tcp_bus
        if isinstance(service, TCPService):
            tcp_bus.tcp_host = service
        if isinstance(service, HTTPService):
            tcp_bus.http_host = service
        service.tcp_bus = tcp_bus
        # service.pubsub_bus = pubsub_bus

    @classmethod
    def _setup_logging(cls):
        identifier = '{}'.format(cls.service_name)
        setup_logging(identifier)
        if cls._smtp_handler:
            logger = logging.getLogger()
            logger.addHandler(cls._smtp_handler)
        Stats.service_name = cls.service_name
        Aggregator._service_name = cls.service_name
        Aggregator.periodic_aggregated_stats_logger()



import copy
import importlib
import json
import logging
import os

from trellio.services import TCPService, HTTPService
from ..utils.log_handlers import BufferingSMTPHandler

logger = logging.getLogger(__name__)

GLOBAL_CONFIG = {
    "RONIN": False,
    "HOST_NAME": "",
    "ADMIN_EMAILS": [],
    "SERVICE_NAME": "",
    "SERVICE_VERSION": "",
    "REGISTRY_HOST": "",
    "REGISTRY_PORT": "",
    "REDIS_HOST": "",
    "REDIS_PORT": "",
    "HTTP_HOST": "",
    "TCP_HOST": "",
    "HTTP_PORT": "",
    "TCP_PORT": "",
    "SIGNALS": {},
    "MIDDLEWARES": [],
    "APPS": [],
    "DATABASE_SETTINGS": {
        "database": "",
        "user": "",
        "password": "",
        "host": "",
        "port": ""
    },
    "SMTP_SETTINGS": {}
}


class InvalidConfigurationError(Exception):
    pass


class ConfigHandler:
    smtp_host = 'SMTP_HOST'
    smtp_user = 'SMTP_USER'
    smtp_port = 'SMTP_PORT'
    smtp_password = 'SMTP_PASSWORD'
    admin_emails = 'ADMIN_EMAILS'
    middleware_key = 'MIDDLEWARES'
    signal_key = 'SIGNALS'
    service_name_key = 'SERVICE_NAME'
    host_name_key = 'HOST_NAME'
    service_version_key = 'SERVICE_VERSION'
    reg_host_key = "REGISTRY_HOST"
    reg_port_key = "REGISTRY_PORT"
    redis_host_key = "REDIS_HOST"
    redis_port_key = "REDIS_PORT"
    http_host_key = "HTTP_HOST"
    tcp_host_key = "TCP_HOST"
    http_port_key = "HTTP_PORT"
    tcp_port_key = "TCP_PORT"
    database_key = 'DATABASE_SETTINGS'
    ronin_key = "RONIN"
    smtp_key = 'SMTP_SETTINGS'
    apps_key = 'APPS'

    # service_path_key = "SERVICE_PATH"

    def __init__(self, host_class):
        self.settings = None
        self.host = host_class

    @property
    def service_name(self):
        return self.settings[self.service_name_key]

    def get_tcp_clients(self):
        from trellio.services import TCPServiceClient
        tcp_clients = self.inheritors(TCPServiceClient)
        return tcp_clients

    def get_http_clients(self):
        from trellio.services import HTTPServiceClient
        http_clients = self.inheritors(HTTPServiceClient)
        return http_clients

    def get_subscribers(self):
        from trellio.pubsub import Subscriber
        subscriber_classes = self.inheritors(Subscriber)
        subscribers = []
        for subs in subscriber_classes:
            s = subs()
            s.pubsub_host = self.settings[self.redis_host_key]
            s.pubsub_port = self.settings[self.redis_port_key]
            subscribers.append(s)
        return subscribers

    def configure_host(self, host):
        host.configure(
            host_name=self.settings[self.host_name_key],
            service_name=self.settings[self.service_name_key],
            service_version=self.settings[self.service_version_key],
            http_host=self.settings[self.http_host_key],
            http_port=self.settings[self.http_port_key],
            tcp_host=self.settings[self.tcp_host_key],
            tcp_port=self.settings[self.tcp_port_key],
            registry_host=self.settings[self.reg_host_key],
            registry_port=self.settings[self.reg_port_key],
            pubsub_host=self.settings[self.redis_host_key],
            pubsub_port=self.settings[self.reg_port_key],
            ronin=self.settings[self.ronin_key]
        )

    def setup_host(self):
        host = self.host
        self.configure_host(host)
        publisher = self.get_publisher()
        subscribers = self.get_subscribers()
        if publisher:
            host.attach_publisher(publisher)
        if subscribers:
            host.attach_subscribers(subscribers)

        http_service = self.get_http_service()
        tcp_service = self.get_tcp_service()
        tcp_clients = self.get_tcp_clients()
        http_clients = self.get_http_clients()
        http_views = self.get_http_views()
        tcp_views = self.get_tcp_views()

        if not http_service:
            http_service = HTTPService(host.service_name, host.service_version, host.http_host, host.http_port)

        if not tcp_service:
            tcp_service = TCPService(host.service_name, host.service_version, host.tcp_host, host.tcp_port)

        self.enable_signals()
        self.enable_middlewares(http_service=http_service, http_views=http_views)

        if http_service:
            # self.register_http_views(http_service)
            host.attach_service(http_service)
            http_service.clients = [i() for i in http_clients + tcp_clients]
            # self.register_tcp_views(tcp_service)

        host.attach_service(tcp_service)

        if http_service:
            tcp_service.clients = http_service.clients

        if http_views:
            host.attach_http_views(http_views)
            for view_inst in host.get_tcp_views():
                pass

        if tcp_views:
            host.attach_tcp_views(tcp_views)
            _tcp_service = host.get_tcp_service()
            _tcp_service.tcp_views = host._tcp_views

        host._smtp_handler = self.get_smtp_logging_handler()

    def get_database_settings(self):
        return self.settings[self.database_key]

    def set_config(self, config_path):
        settings = None
        with open(config_path) as f:
            settings = json.load(f)
        new_settings = copy.deepcopy(GLOBAL_CONFIG)
        new_settings.update(settings)
        self.settings = new_settings
        parent_dir = os.getcwd().split('/')[-1]
        client_path = parent_dir + '.clients'
        service_path = parent_dir + '.service'

        try:
            importlib.import_module(client_path)
        except:
            logger.warning('No clients found')

        service_imported = True
        service_exception = None
        try:
            importlib.import_module(service_path)
        except Exception as e:
            service_imported = False
            service_exception = e.__traceback__

        if self.settings.get(self.apps_key):
            apps = self.settings[self.apps_key]
            for app in apps:
                views_path = parent_dir + '.{}.views'.format(app)
                try:
                    importlib.import_module(views_path)
                except Exception as e:
                    print(e.__traceback__)
        else:
            if not service_imported:
                print(service_exception)

    def get_smtp_logging_handler(self):
        if self.settings.get(self.smtp_key):
            keys = ["smtp_host", "smtp_port", "smtp_user", "smtp_password"]
            setting_keys = self.settings[self.smtp_key].keys()
            missing_keys = list(filter(lambda x: x not in setting_keys, keys))
            if not missing_keys:
                handler = BufferingSMTPHandler(mailhost=self.settings[self.smtp_key]['smtp_host'],
                                               mailport=self.settings[self.smtp_key]['smtp_port'],
                                               fromaddr=self.settings[self.smtp_key]['smtp_user'],
                                               toaddrs=self.settings[self.admin_emails],
                                               subject='Error {} {}:{}'.format(self.settings[self.host_name_key],
                                                                               self.settings[
                                                                                   self.service_name_key].upper(),
                                                                               self.settings[self.service_version_key]),
                                               capacity=1,
                                               password=self.settings[self.smtp_key]['smtp_password'])
                handler.setLevel(logging.ERROR)
                if not self.settings[self.ronin_key]:
                    return handler

    def get_http_service(self):
        from trellio.services import HTTPService
        http_service = None
        if HTTPService.__subclasses__():
            service_sub_class = HTTPService.__subclasses__()[0]

            http_service = service_sub_class(self.settings[self.service_name_key],
                                             self.settings[self.service_version_key],
                                             self.settings[self.http_host_key],
                                             self.settings[self.http_port_key])
        return http_service

    def get_tcp_service(self):
        from trellio.services import TCPService
        tcp_service = None
        if TCPService.__subclasses__():
            service_sub_class = TCPService.__subclasses__()[0]
            tcp_service = service_sub_class(self.settings[self.service_name_key],
                                            self.settings[self.service_version_key],
                                            self.settings[self.tcp_host_key],
                                            self.settings[self.tcp_port_key])
        return tcp_service

    def get_publisher(self):
        from trellio.pubsub import Publisher
        publisher = None
        if Publisher.__subclasses__():
            publisher_sub_class = Publisher.__subclasses__()[0]
            publisher = publisher_sub_class(self.settings[self.service_name_key],
                                            self.settings[self.service_version_key],
                                            self.settings[self.redis_host_key],
                                            self.settings[self.redis_port_key])
        return publisher

    def get_http_views(self):
        from trellio.views import HTTPView
        return self.inheritors(HTTPView)

    def get_tcp_views(self):
        from trellio.views import TCPView
        return self.inheritors(TCPView)

    def import_class_from_path(self, path):
        broken = path.split('.')
        class_name = broken[-1]
        module_name = '.'.join(broken[:-1])
        module = importlib.import_module(module_name)
        class_value = getattr(module, class_name)
        return module, class_value

    def enable_middlewares(self, http_service=None, http_views=()):
        middlewares = self.settings[self.middleware_key] or []
        middle_cls = []
        for i in middlewares:
            module, class_value = self.import_class_from_path(i)
            if not class_value:
                raise InvalidConfigurationError
            else:
                middle_cls.append(class_value())

        if http_service:
            http_service.middlewares = middle_cls
        for view in http_views:
            view.middlewares = middle_cls

    def enable_signals(self):
        '''
        e.g signal_dict = {signal_path:signal_receiver_path_list, ....}
        :return:
        '''
        signal_dict = self.settings[self.signal_key] or {}
        for i in signal_dict.keys():
            sig_module, signal_class = self.import_class_from_path(i)
            for j in signal_dict[i]:
                recv_module, recv_coro = self.import_class_from_path(j)
                signal_class.register(recv_coro)  # registering reciever

    @staticmethod
    def inheritors(klass):
        subclasses = set()
        work = [klass]
        while work:
            parent = work.pop()
            for child in parent.__subclasses__():
                if child not in subclasses:
                    subclasses.add(child)
                    work.append(child)
        return list(subclasses)



from aiohttp.web import Request as Req, Response as Res


class Request(Req):
    """
    Wraps the aiohttp request object to hide it from user
    """
    pass


class Response(Res):
    """
    Wraps the aiohttp response object to hide it from user
    """
    pass



class ManagementCommand:
    name = ''

    def run(self):
        raise NotImplementedError


class ManagementRegistry:
    _management_reg = {}

    @classmethod
    def register(cls, command_cls):
        if issubclass(command_cls, ManagementCommand):
            cls._management_reg[command_cls.name] = command_cls

    @classmethod
    def get(cls, name):
        return cls._management_reg.get(name)
